########################################################################
# $Header: /var/local/cvsroot/4Suite/Ft/Xml/XPath/__init__.py,v 1.23 2004/06/24 07:27:23 mbrown Exp $
"""
4XPath initialization and principal functions

Copyright 2003 Fourthought, Inc. (USA).
Detailed license and copyright information: http://4suite.org/COPYRIGHT
Project home, documentation, distributions: http://4suite.org/
"""

NAMESPACE_NODE = 10000
FT_OLD_EXT_NAMESPACE = 'http://xmlns.4suite.org/xpath/extensions'
FT_EXT_NAMESPACE = 'http://xmlns.4suite.org/ext'

import types
from xml.dom import Node

from Ft.Lib import boolean
from Ft.Lib.number import nan as NaN
from Ft.Lib.number import inf as Inf

# -- XPath exceptions --------------------------------------------------

from Ft import FtException

class XPathException(FtException):
    """
    Base class for exceptions specific to XPath processing
    """
    def __init__(self, errorCode, messages, args):
        FtException.__init__(self, errorCode, messages, args)


class CompiletimeException(XPathException):
    """
    The exception raised when an error is encountered during the
    parsing or compilation of an XPath expression.
    """
    #errorCodes:
    # internal/unexpected errors
    INTERNAL = 1
    # other expression compile-time errors
    SYNTAX = 2

    def __init__(self, errorCode, *args):
        # import here to avoid circularity
        import MessageSource
        XPathException.__init__(self, errorCode,
                                MessageSource.ERROR_COMPILETIME, args)
        return


class RuntimeException(XPathException):
    """
    The exception raised when an error is encountered during the
    parsing or evaluation of an XPath expression.
    """
    #errorCodes:
    # internal/unexpected errors
    INTERNAL = 1
    # other expression evaluation (run-time) errors
    NO_CONTEXT         = 10
    UNDEFINED_VARIABLE = 100
    UNDEFINED_PREFIX   = 101
    UNDEFINED_FUNCTION = 102
    WRONG_ARGUMENTS    = 200
    ARGCOUNT_NONE      = 201
    ARGCOUNT_ATLEAST   = 202
    ARGCOUNT_EXACT     = 203
    ARGCOUNT_ATMOST    = 204

    def __init__(self, errorCode, *args):
        # import here to avoid circularity
        import MessageSource
        XPathException.__init__(self, errorCode,
                                MessageSource.ERROR_RUNTIME, args)
        return



# -- Python primitive types used for XPath object types ----------------

g_xpathPrimitiveTypes = [
    types.StringType,
    types.UnicodeType,
    types.IntType,
    types.LongType,
    types.FloatType,
    types.ListType,
    boolean.BooleanType,
    ]

# -- DOM node types used for XPath node types --------------------------

g_xpathRecognizedNodes = [
        Node.ELEMENT_NODE,
        Node.ATTRIBUTE_NODE,
        Node.TEXT_NODE,
        Node.DOCUMENT_NODE,
        Node.PROCESSING_INSTRUCTION_NODE,
        Node.COMMENT_NODE
        ]

# -- XPath API ---------------------------------------------------------

def Evaluate(expr, contextNode=None, context=None):
    """
    Evaluates the given XPath expression.

    Two arguments are required: the expression (as a string or compiled
    expression object), and a context. The context can be given as a
    Domlette node via the 'contextNode' named argument, or can be given as
    an Ft.Xml.XPath.Context.Context object via the 'context' named
    argument.

    If namespace bindings or variable bindings are needed, use a
    Context object. If extension functions are needed, either use a
    Context object, or set the EXTMODULES environment variable to be a
    ':'-separated list of names of Python modules that implement
    extension functions.

    The return value will be one of the following:
    node-set: list of Domlette node objects (xml.dom.Node based);
    string: Unicode string type;
    number: float type;
    boolean: Ft.Lib.boolean C extension object;
    or a non-XPath object (i.e. as returned by an extension function).
    """
    import os
    if os.environ.has_key('EXTMODULES'):
        ext_modules = os.environ["EXTMODULES"].split(':')
    else:
        ext_modules = []

    if contextNode and context:
        con = context.clone()
        con.node = contextNode
    elif context:
        con = context
    elif contextNode:
        #contextNode should be a node, not a context obj,
        #but this is a common error.  Be forgiving?
        if isinstance(contextNode, Context.Context):
            con = contextNode
        else:
            con = Context.Context(contextNode, 0, 0, extModuleList=ext_modules)
    else:
        raise RuntimeException(RuntimeException.NO_CONTEXT)

    if hasattr(expr, "evaluate"):
        retval = expr.evaluate(con)
    else:
        retval = parser.new().parse(expr).evaluate(con)
    return retval


def Compile(expr):
    """
    Given an XPath expression as a string, returns an object that allows
    an evaluation engine to operate on the expression efficiently.
    This "compiled" expression object can be passed to the Evaluate
    function instead of a string, in order to shorten the amount of time
    needed to evaluate the expression.
    """
    if type(expr) not in [types.StringType, types.UnicodeType]:
        raise TypeError("Expected string, found %s" % type(expr))
    try:
        return parser.new().parse(expr)
    except SyntaxError, error:
        raise CompiletimeException(CompiletimeException.SYNTAX, 0, 0, str(error))
    except:
        import traceback, cStringIO
        stream = cStringIO.StringIO()
        traceback.print_exc(None, stream)
        raise CompiletimeException(CompiletimeException.INTERNAL, stream.getvalue())


def CreateContext(contextNode):
    return Context.Context(contextNode, 0, 0)

#Allow access to the NormalizeNode function
from Util import NormalizeNode


# -- Additional setup --------------------------------------------------

import Context

try:
    import XPathParserc
    parser = XPathParserc
except ImportError:
    import XPathParser
    parser = XPathParser

